"""$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/__init__.py $
$Id: __init__.py 23965 2004-04-12 20:49:21Z dbinger $
Copyright (c) Corporation for National Research Initiatives 2004. All Rights Reserved.
"""
